/***************************************************************************
    kcdspeed.cpp  -  CD-ROM speed setting applet
    ------------------------------------------------------------------------
    begin                : Thu Jun  6 2002
    copyright            : (C) 2002, 2005 by Tomas Horsky
    email                : tomas@riadoklan.sk
 ***************************************************************************/

#include <kglobal.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kaboutdata.h>
#include <kaboutapplication.h>
#include <kcombobox.h>
#include <kpushbutton.h>
#include <kpopupmenu.h>
#include <kiconloader.h>
#include <kio/global.h>
#include <kautomount.h>
#include <kprocess.h>
#include <qtooltip.h>
#include "cdromdevice.h"
#include "kcdssettings.h"
#include "kcdspreferences.h"

#include "kcdspeed.h"

extern "C" {

KPanelApplet *init(QWidget *parent, const QString &configFile)
{
	KGlobal::locale()->insertCatalogue("kcdspeed");
	return new KCdSpeed(configFile, KPanelApplet::Normal,
	                    KPanelApplet::About | KPanelApplet::Preferences,
	                    parent, "kcdspeed");
}

}

#define MBOX_CAPTION i18n("KCdSpeed applet")
#define MBOX_ERROR_CAPTION i18n("KCdSpeed applet error")

#define UNKNOWN_SPEED_TEXT "?"

//Minimum applet size (width or height) for square layout
#define MIN_SIZE_FOR_2X2 32
//Minimum combobox width (when we have fixed height)
#define MIN_COMBO_WIDTH 44
//Maximum combobox height (when we have fixed width)
#define MAX_COMBO_HEIGHT 20

KCdSpeed::KCdSpeed(const QString &configFile, Type t, int actions, QWidget *parent, const char *name):
	KPanelApplet(configFile, t, actions, parent, name),
	_preferences(0)
{
	//create settings object
	_settings=new KcdsSettings(config());
	
	//create CD-ROM device object
	_cdrom=new CdromDevice(_settings);
	connect(_cdrom, SIGNAL(openDeviceError(const char *)), SLOT(slotOpenDeviceError(const char *)));
	
	connect(_cdrom, SIGNAL(speedSelected(int)), SLOT(slotSpeedSelected(int)));
	connect(_cdrom, SIGNAL(speedSelectError(int)), SLOT(slotSpeedSelectError(int)));
	
	connect(_cdrom, SIGNAL(trayEjected()), SLOT(slotTrayEjected()));
	connect(_cdrom, SIGNAL(trayEjectError()), SLOT(slotTrayEjectError()));
	
	connect(_cdrom, SIGNAL(trayClosed()), SLOT(slotTrayClosed()));
	connect(_cdrom, SIGNAL(trayCloseError()), SLOT(slotTrayCloseError()));
	
	//construct applet gui
	_speeds=new KComboBox(this, "_speeds");
	_speeds->installEventFilter(this);	//to catch right mouse button clicks
	QToolTip::add(_speeds, i18n("Selects new CD-ROM speed"));
	connect(_speeds, SIGNAL(activated(const QString &)), SLOT(slotSelectSpeed(const QString &)));
	
	_eject=new KPushButton(this, "_eject");
	_eject->setText("E");
	QToolTip::add(_eject, i18n("Ejects CD-ROM tray"));
	connect(_eject, SIGNAL(clicked()), SLOT(slotEjectTray()));
	
	_close=new KPushButton(this, "_close");
	_close->setText("C");
	QToolTip::add(_close, i18n("Closes CD-ROM tray"));
	connect(_close, SIGNAL(clicked()), SLOT(slotCloseTray()));
	
	updateWidgets();
	
	//what speed set at startup ?
	if(_cdrom->persistentOpen()) {
		switch(_settings->startup()) {
			case KcdsSettings::Minimum: {
				int item=-1;
				int min=1000;
				for(int i=0; i<_speeds->count(); i++) {
					int sp=CdromDevice::speedToInt(_speeds->text(i));
					if(sp>0 && sp<min) {
						item=i;
						min=sp;
					}
				}
				_speeds->setCurrentItem(item);
				slotSelectSpeed(_speeds->text(item));
				break;
			}
			case KcdsSettings::Maximum: {
				int item=-1;
				int max=0;
				for(int i=0; i<_speeds->count(); i++) {
					int sp=CdromDevice::speedToInt(_speeds->text(i));
					if(sp>0 && sp>max) {
						item=i;
						max=sp;
					}
				}
				_speeds->setCurrentItem(item);
				slotSelectSpeed(_speeds->text(item));
				break;
			}
			case KcdsSettings::Last: {
				if(_settings->lastSpeed()!="") {
					QString speed=_settings->lastSpeed();
					if(findSpeed(speed))	//if there is that speed in combobox
						slotSelectSpeed(speed);	//change speed to it
				}
				break;
			}
			default: {
				break;
			}
		}
		_cdrom->persistentClose();
	}
}

KCdSpeed::~KCdSpeed()
{
	if(_preferences)
		delete _preferences;
	delete _cdrom;
	delete _settings;
}

int KCdSpeed::widthForHeight(int height) const
{
	if(height<MIN_SIZE_FOR_2X2)
		return MIN_COMBO_WIDTH*2;	//line layout
	return height<MIN_COMBO_WIDTH? MIN_COMBO_WIDTH: height;	//square layout
}

int KCdSpeed::heightForWidth(int width) const
{
	if(width<MIN_SIZE_FOR_2X2)
		return MAX_COMBO_HEIGHT*3;	//line layout
	return width>MAX_COMBO_HEIGHT*2? MAX_COMBO_HEIGHT*2: width;	//square layout
}

void KCdSpeed::slotOpenDeviceError(const char *device)
{
	QString msg=i18n("Could not open CD-ROM device '");
	msg+=device;
	msg+="'.";
	KMessageBox::sorry(this, msg, MBOX_ERROR_CAPTION);
}

void KCdSpeed::slotSelectSpeed(const QString &speed)
{
	int sp=CdromDevice::speedToInt(speed);
	if(sp==-1) {	//if user clicked at "?", we ignore it
		_speeds->setCurrentItem(_prevItem);
		return;
	}
	_cdrom->slotSelectSpeed(sp);
}

void KCdSpeed::slotSpeedSelected(int s)
{
	removeUnknownSpeed();
	_prevItem=_speeds->currentItem();
	QString speed=CdromDevice::intToSpeed(s);
	_settings->setLastSpeed(speed);
	if(_settings->confirmSpeed())
		KMessageBox::information(this, i18n("CD-ROM speed has been changed to ") +speed +".", MBOX_CAPTION);
}

void KCdSpeed::slotSpeedSelectError(int s)
{
	if(!_settings->ignoreErrors())
		KMessageBox::sorry(this, i18n("Could not change CD-ROM speed to ") +CdromDevice::intToSpeed(s) +".", MBOX_ERROR_CAPTION);
	_speeds->setCurrentItem(_prevItem);
}

void KCdSpeed::slotEjectTray()
{
	_eject->setEnabled(false);
	if(_settings->autoUnmount() && KIO::findDeviceMountPoint(_settings->device())!="") {
		slotUnmount(true);	//this will eject it after unmount succeeded
		return;
	}
	_cdrom->slotEjectTray();
}

void KCdSpeed::slotTrayEjected()
{
	if(!_settings->lockButtons())
		_eject->setEnabled(true);
}

void KCdSpeed::slotTrayEjectError()
{
	if(!_settings->lockButtons())
		_eject->setEnabled(true);
	KMessageBox::sorry(this, i18n("Could not eject CD-ROM tray."), MBOX_ERROR_CAPTION);
}

void KCdSpeed::slotCloseTray()
{
	_close->setEnabled(false);
	_cdrom->slotCloseTray();
}

void KCdSpeed::slotTrayClosed()
{
	if(!_settings->lockButtons())
		_close->setEnabled(true);
	if(_settings->autoMount() && KIO::findDeviceMountPoint(_settings->device())=="")
		slotMount();
}

void KCdSpeed::slotTrayCloseError()
{
	if(!_settings->lockButtons())
		_close->setEnabled(true);
	KMessageBox::sorry(this, i18n("Could not close CD-ROM tray."), MBOX_ERROR_CAPTION);
}

void KCdSpeed::slotMount()
{
	new KAutoMount(true, QString::null, _settings->device(), _settings->customMountPoint()? _settings->mountPoint(): QString::null, QString::null, _settings->openFilemanager());
}

void KCdSpeed::slotUnmount(bool eject)
{
	KAutoUnmount *unm=new KAutoUnmount(KIO::findDeviceMountPoint(_settings->device()), QString::null);
	if(eject) {
		connect(unm, SIGNAL(finished()), _cdrom, SLOT(slotEjectTray()));
		connect(unm, SIGNAL(error()), _cdrom, SLOT(slotEjectTray()));	//to enable "E" button again if unmount failed
	}
}

void KCdSpeed::slotOpenFilemanager()
{
	KProcess proc;
	proc << "kfmclient" << "openURL" << KIO::findDeviceMountPoint(_settings->device());
	proc.start(KProcess::DontCare);
}

void KCdSpeed::slotOpenPreferences()
{
	if(_preferences) {
		_preferences->raise();
		return;
	}
	_preferences=new KcdsPreferences(this, 0, false, WDestructiveClose);
	_preferences->readSettings(_settings);
	connect(_preferences, SIGNAL(applyButtonPressed()), SLOT(slotPreferencesApply()));
	connect(_preferences, SIGNAL(destroyed()), SLOT(slotPreferencesDeleted()));
	_preferences->show();
}

void KCdSpeed::slotPreferencesApply()
{
	//get new settings from dialog
	_preferences->writeSettings(_settings);
	//write new settings to config
	_settings->writeConfig();
	
	updateWidgets();
}

void KCdSpeed::slotPreferencesDeleted()
{
	_preferences=0;
}

void KCdSpeed::slotShowCapabilities()
{
	if(!_cdrom->persistentOpen())
		return;
	QString mesg(i18n("CD-ROM device says, that it\n"));
	mesg+=_cdrom->canSelectSpeed()? i18n("- can "): i18n("- can NOT "); mesg+=i18n("select speed,\n");
	mesg+=_cdrom->canEjectTray()? i18n("- can "): i18n("- can NOT "); mesg+=i18n("eject tray,\n");
	mesg+=_cdrom->canCloseTray()? i18n("- can "): i18n("- can NOT "); mesg+=i18n("close tray,\n");
	mesg+=_cdrom->canLockTray()? i18n("- can "): i18n("- can NOT "); mesg+=i18n("lock tray.\n");
	_cdrom->persistentClose();
	KMessageBox::information(this, mesg, MBOX_CAPTION);
}

void KCdSpeed::slotToggleLockButtons()
{
	if(_settings->lockButtons()) {
		_settings->setLockButtons(false);
		_eject->setEnabled(true);
		_close->setEnabled(true);
	} else {
		_settings->setLockButtons(true);
		_eject->setEnabled(false);
		_close->setEnabled(false);
	}
}

void KCdSpeed::about()
{
	KAboutData aboutData("kcdspeed", I18N_NOOP("KCdSpeed"),
	                     VERSION, I18N_NOOP("CD-ROM speed setting applet"),
	                     KAboutData::License_GPL, "(c) 2002, 2005 Tomas Horsky", 0, 0,
	                     "tomas@riadoklan.sk");
	aboutData.addAuthor("Tomas Horsky", 0, "tomas@riadoklan.sk");
	KAboutApplication about(&aboutData, this);
	about.exec();
}

void KCdSpeed::updateWidgets()
{
	if(!_cdrom->testValidDevice(true)) {
		_speeds->setEnabled(false);
		_eject->setEnabled(false);
		_close->setEnabled(false);
		showUnknownSpeed();
	} else {
		_speeds->setEnabled(true);
		if(_settings->lockButtons()) {
			_eject->setEnabled(false);
			_close->setEnabled(false);
		} else {
			_eject->setEnabled(true);
			_close->setEnabled(true);
		}
		if(availSpeedsChanged()) {
			//remember current text in combobox
			QString currSpeed=_speeds->currentText();
			_speeds->clear();
			_speeds->insertStringList(_settings->availSpeeds());
			//now find currSpeed in combobox, and if it is still there, show it
			if(!findSpeed(currSpeed))
				showUnknownSpeed();
		}
	}
}

void KCdSpeed::showUnknownSpeed()
{
	if(_speeds->text(0)==UNKNOWN_SPEED_TEXT)
		return;
	_speeds->insertItem(UNKNOWN_SPEED_TEXT, 0);
	_speeds->setCurrentItem(0);
	_prevItem=_speeds->currentItem();
}

void KCdSpeed::removeUnknownSpeed()
{
	if(_speeds->text(0)==UNKNOWN_SPEED_TEXT) {
		int currItem=_speeds->currentItem();
		_speeds->removeItem(0);
		if(currItem>0)
			_speeds->setCurrentItem(currItem -1);
	}
}

bool KCdSpeed::findSpeed(QString &speed)
{
	for(int i=0; i<_speeds->count(); i++)
		if(_speeds->text(i)==speed) {
			_speeds->setCurrentItem(i);
			return true;
		}
	return false;
}

bool KCdSpeed::availSpeedsChanged()
{
	QStringList cs;
	for(int i=0; i<_speeds->count(); i++)
		cs+=_speeds->text(i);
	if(cs==_settings->availSpeeds())
		return false;
	return true;
}

void KCdSpeed::resizeEvent(QResizeEvent *)
{
	switch(position()) {
		case KPanelApplet::pLeft:
		case KPanelApplet::pRight:
			if(width()<MIN_SIZE_FOR_2X2) {	//line layout
				_speeds->setFixedSize(width(), height()/3);
				_eject->setFixedSize(width(), height()/3);
				_close->setFixedSize(width(), height()/3);
				_eject->move(0, height()/3);
				_close->move(0, height()*2/3);
			} else {	//square layout
				_speeds->setFixedSize(width(), height()/2);
				_eject->setFixedSize(width()/2, height()/2);
				_close->setFixedSize(width()/2, height()/2);
				_eject->move(0, height()/2);
				_close->move(width()/2, height()/2);
			}
			break;
		case KPanelApplet::pTop:
		case KPanelApplet::pBottom:
			if(height()<MIN_SIZE_FOR_2X2) {	//line layout
				_speeds->setFixedSize(width()/2, height());
				_eject->setFixedSize(width()/4, height());
				_close->setFixedSize(width()/4, height());
				_eject->move(width()/2, 0);
				_close->move(width()*3/4, 0);
			} else {	//square layout
				_speeds->setFixedSize(width(), height()/2);
				_eject->setFixedSize(width()/2, height()/2);
				_close->setFixedSize(width()/2, height()/2);
				_eject->move(0, height()/2);
				_close->move(width()/2, height()/2);
			}
			break;
	}
}

void KCdSpeed::mousePressEvent(QMouseEvent *event)
{
	if(event->button()==RightButton) {
		KPopupMenu *_menu=new KPopupMenu(this);
		_menu->setCheckable(true);
		KIconLoader icons("kcdspeed");
		_menu->insertTitle(icons.loadIcon("cdrom_unmount", KIcon::Small), i18n("KCdSpeed"));
		_menu->insertItem(icons.loadIcon("configure", KIcon::Small), i18n("&Preferences"), this, SLOT(slotOpenPreferences()));
		int idCapa=_menu->insertItem(i18n("&Capabilities"), this, SLOT(slotShowCapabilities()));
		_menu->insertSeparator();
		
		int idLock=_menu->insertItem(i18n("&Lock buttons"), this, SLOT(slotToggleLockButtons()));
		_menu->setItemChecked(idLock, _settings->lockButtons());
		_menu->insertSeparator();
		
		int idOFm=_menu->insertItem(i18n("&Open filemanager"), this, SLOT(slotOpenFilemanager()));
		int idMnt;
		if(KIO::findDeviceMountPoint(_settings->device())=="") {	//is it mounted ?
			_menu->setItemEnabled(idOFm, false);	//no
			idMnt=_menu->insertItem(i18n("&Mount"), this, SLOT(slotMount()));
		} else
			idMnt=_menu->insertItem(i18n("&Unmount"), this, SLOT(slotUnmount()));	//yes
		
		if(!_cdrom->isValidDevice()) {
			_menu->setItemEnabled(idLock, false);
			_menu->setItemEnabled(idCapa, false);
			_menu->setItemEnabled(idMnt, false);
		}
		_menu->exec(mapToGlobal(event->pos()));
		delete _menu;
	}
}

bool KCdSpeed::eventFilter(QObject *watched, QEvent *event)
{
	if(event->type()==QEvent::MouseButtonPress) {
		QMouseEvent *mevent=(QMouseEvent *)event;
		if(mevent->button()==RightButton) {
			mousePressEvent(mevent);
			return true;
		}
	}
	return KPanelApplet::eventFilter(watched, event);
}
